#include "evaluation.h"
#include <libxml/xmlreader.h>
//  #include "thirdparty/CVML.h"

PerformanceEvaluator::PerformanceEvaluator()
{
	frames_evaluated = 0;
	frames_successfully_tracked = 0;
	overlapping_bboxes = 0.0;

	// auto id generation
	id = -1;
	// for debugging, hardcode id:
	// id = 0;
}

void PerformanceEvaluator::addFrameInformation(TrackingObject *object)
{
	readNextFrame();

	CvRect obj_bbox = object->getBoundingBox();
	// calculate overlapping bounding boxes
	long area1 = curr_gt_bbox.width * curr_gt_bbox.height;
	long area2 = obj_bbox.width * obj_bbox.height;
	CvRect r1 = curr_gt_bbox, r2 = obj_bbox;

	// calc overlapping rect of object over real bbox
	if(obj_bbox.x > r1.x)
	{
		r1.x = obj_bbox.x;
		r1.width -= (obj_bbox.x-r1.x);
	}
	if(obj_bbox.y > r1.y)
	{
		r1.y = obj_bbox.y;
		r1.height -= (obj_bbox.y-r1.y);
	}
	if(obj_bbox.width < r1.width)
		r1.width = obj_bbox.width;
	if(obj_bbox.height < r1.height)
		r1.height = obj_bbox.height;

	// calc overlapping rect of bbox over object
	if(curr_gt_bbox.x > r2.x)
	{
		r2.x = curr_gt_bbox.x;
		r2.width -= (curr_gt_bbox.x-r2.x);
	}
	if(curr_gt_bbox.y > r2.y)
	{
		r2.y = curr_gt_bbox.y;
		r2.height -= (curr_gt_bbox.y-r2.y);
	}
	if(curr_gt_bbox.width < r2.width)
		r2.width = curr_gt_bbox.width;
	if(curr_gt_bbox.height < r2.height)
		r2.height = curr_gt_bbox.height;

	long area3 = r1.width * r1.height;
	long area4 = r2.width * r2.height;

	float perc1 = (float)area3/area1;
	float perc2 = (float)area4/area2;

	float min_perc = perc1; if (perc2 < min_perc) min_perc = perc2;

	printf("  overlap: %3.2f%%\n", min_perc*100);

	frames_evaluated++;
	// at least 30% overlapping bounding boxes
	if(min_perc > 0.3)
		frames_successfully_tracked++;

	overlapping_bboxes += min_perc;
}



MusclePerformanceEvaluator::MusclePerformanceEvaluator(const char* ground_truth_file, int skip_frames)
	: PerformanceEvaluator()
{
	file = fopen(ground_truth_file, "r");
	if(!file)
		fprintf(stderr, "File %s not readable, evaluation not possible.\n", ground_truth_file);

	while(skip_frames--)
		readNextFrame();
}

MusclePerformanceEvaluator::~MusclePerformanceEvaluator()
{
	if(file)
		fclose(file);
}

/* file format:
00010
id 1: 230 198 bb: 211 249 179 217 ty: 1 39
id 2: 105 60 bb: 86 124 41 79 ty: 1 39
*/
bool MusclePerformanceEvaluator::readNextFrame()
{
	char line[1024];
	while(fgets(line, 1024, file) != NULL)
	{
		char *endptr;
		long frame = strtol(line, &endptr, 10);
		if (endptr != line) // numbers found
		{
			int i,c1,c2,bb1,bb2,bb3,bb4,ty1,ty2;
			while(fgets(line, 1024, file) != NULL)
			{
				int matched = sscanf(line, "id %d: %d %d bb: %d %d %d %d ty: %d %d", &i,&c1,&c2,&bb1,&bb2,&bb3,&bb4,&ty1,&ty2);
				if (matched < 7)
					return false;

				if (id == -1) // uninitialized, find id
				{
					if(point_of_id.x > bb1 && point_of_id.x < bb2 && point_of_id.y > bb3 && point_of_id.y < bb4)
						id = i;
					else
						continue;
				}

				if(i == id)
				{
					fprintf(stderr, "Evaluator read frame %ld\n  bounding box: %d,%d,%d,%d\n", frame+1, bb1,bb3,bb2,bb4);
					curr_gt_bbox = cvRect(bb1, bb3, bb2-bb1, bb4-bb3);
					curr_gt_center = cvPoint(c1,c2);
					return true;
				}
			}
		}
	}
	return false;
}

// fileformat: XML - verboooooose ;)
bool CaviarPerformanceEvaluator::readNextFrame()
{
	const xmlChar *name, *value;

	int i = -1;
	int x,y,width,height;

	int ret = xmlTextReaderRead(reader);
	bool tagclosed = true;
	while (ret == 1)
	{
		name = xmlTextReaderConstName(reader);
		if (name == NULL)
			name = BAD_CAST "--";

		value = xmlTextReaderConstValue(reader);

//		printf("%d %d %s %d %d\n", 
//				xmlTextReaderDepth(reader),
//				xmlTextReaderNodeType(reader),
//				name,
//				xmlTextReaderIsEmptyElement(reader),
//				xmlTextReaderHasValue(reader));

		if(!xmlStrcmp(name, BAD_CAST "frame"))
		{
			tagclosed = !tagclosed;
			/*xmlChar* val =*/ xmlTextReaderGetAttribute(reader, BAD_CAST "number");
			if(tagclosed) // </frame>
				return true;
		}
		else if(!xmlStrcmp(name, BAD_CAST "object"))
		{
			xmlChar* ic = xmlTextReaderGetAttribute(reader, BAD_CAST "id");
			i = atoi((char*)ic);
			//printf("id: %d\n", id);
		}
		else if(!xmlStrcmp(name, BAD_CAST "box"))
		{
			xmlChar* ch = xmlTextReaderGetAttribute(reader, BAD_CAST "h");
			xmlChar* cw = xmlTextReaderGetAttribute(reader, BAD_CAST "w");
			xmlChar* cx = xmlTextReaderGetAttribute(reader, BAD_CAST "xc");
			xmlChar* cy = xmlTextReaderGetAttribute(reader, BAD_CAST "yc");

			x = atoi((char*)cx);
			y = atoi((char*)cy);
			width = atoi((char*)cw);
			height = atoi((char*)ch);
			x -= width/2; y -= height/2;

			if (id == -1) // uninitialized, find id
			{
				if(point_of_id.x > x && point_of_id.x < x+width && point_of_id.y > y && point_of_id.y < y+height/2)
				{
					id = i;
					printf("id found! %d\n", id);
				}
			}

			if(id == i)
			{
				//printf("BB: %d,%d,%d,%d\n", x,y,width,height);
				curr_gt_center = cvPoint(x + width/2, y + height/2);
				curr_gt_bbox = cvRect(x, y, width, height);
			}

		}
		ret = xmlTextReaderRead(reader);
	}
	if (ret != 0)
		fprintf(stderr, "failed to parse xml file\n");

	return false;
}

CaviarPerformanceEvaluator::CaviarPerformanceEvaluator(const char* ground_truth_file, int skip_frames)
	: PerformanceEvaluator()
{
    reader = xmlReaderForFile(ground_truth_file, NULL, 0);
    if (reader == NULL)
	{
        fprintf(stderr, "Unable to open %s\n", ground_truth_file);
		return;
    }
	while(skip_frames--)
		readNextFrame();
}

CaviarPerformanceEvaluator::~CaviarPerformanceEvaluator()
{
	if(reader)
	{
        xmlFreeTextReader(reader);
		/*
		 * Cleanup function for the XML library.
		 */
		xmlCleanupParser();
	}
}
