#include <iostream>
#include <vector>

#include "flow.h"
#include "highgui.h"

// utility function to find median of a vector
extern int get_median(std::vector<int> &vec);
/*{
	std::sort(vec.begin(), vec.end());
	int median_value = 0;
	int median_count = 0;
	int this_count = 1;
	int last_val = INT_MAX;
	for(unsigned int i=0; i < vec.size(); i++)
	{
		// printf("val: %d, this_count: %d\n", vec.at(i), this_count);
		if (vec.at(i) == last_val)
			this_count++;
		else
			this_count = 1;

		if (this_count > median_count)
		{
			median_value = vec.at(i);
			median_count = this_count;
		}
		last_val = vec.at(i);
	}
	return median_value;
}*/


// input is a 8bit gray rgb image
OpticalFlow::OpticalFlow(const IplImage *first_image, const CvPoint2D32f* initial_features, const int initial_feature_count)
{
	curr_pyramid = cvCreateImage( cvGetSize(first_image), 8, 1 );
	prev_pyramid = cvCreateImage( cvGetSize(first_image), 8, 1 );

	num_features = 0;

	if (initial_features)
	{
		char status[MAX_FEATURE_POINTS];
		memset(status, FLOW_STATUS_OK, initial_feature_count);

		setCurrentFeaturePoints(initial_features, status, initial_feature_count);
	}
}

OpticalFlow::~OpticalFlow()
{
	if(curr_pyramid)
		cvReleaseImage(&curr_pyramid);
	if(prev_pyramid)
		cvReleaseImage(&prev_pyramid);
}

void OpticalFlow::setCurrentFeaturePoints(const CvPoint2D32f* points, const char* status, const int num_points)
{
	static bool firsttime = true;
	memcpy(curr_features, points, sizeof(CvPoint2D32f) * num_points);
	memcpy(curr_status, status, sizeof(char) * num_points);
	if (firsttime)
	{
//		memcpy(prev_features, points, sizeof(CvPoint2D32f) * num_points);
		memcpy(guess_features, points, sizeof(CvPoint2D32f) * num_features);
//		memcpy(last_nonoccluded_features, points, sizeof(CvPoint2D32f) * num_points);
	}

	firsttime = false;
	num_features = num_points;
}

// this function also sets the number of found and predicted feature points
// returns the number of 'lost' features, which can't be tracked anymore
int OpticalFlow::correct(CvPoint2D32f* features, TrackingObject* prev)
{
	int lost_features = 0;
	if (!features || !prev)
		return lost_features;

	std::vector<int> vec_x, vec_y;
	for(int i=0; i < num_features; i++) 
	{
		if (curr_status[i] == FLOW_STATUS_OK)
		{
			float x = features[i].x - prev->features[i].x;
			float y = features[i].y - prev->features[i].y;
			int ix = (int)nearbyintf(x);
			int iy = (int)nearbyintf(y);
			vec_x.push_back(ix);
			vec_y.push_back(iy);
		}
	}

	CvPoint median;
	median.x = get_median(vec_x);
	median.y = get_median(vec_y);
	//printf("Median direction: %d,%d of %d good features\n", median.x, median.y, vec_x.size());

	// mark certain feature points as bad
	int num_found_features = 0, num_predicted_features = 0;
	float avg_x = 0.f, avg_y = 0.f;
	for(int i=0; i < num_features; i++) 
	{
		if (curr_status[i] == FLOW_STATUS_OK)// && prev->status[i] == FLOW_STATUS_OK)
		{
			float x = features[i].x - prev->features[i].x;
			float y = features[i].y - prev->features[i].y;

			// points is more than 2 pixel and 20% away of median direction
			float max_deviation = 2.f;
			if( (fabs(x - median.x) > max_deviation && fabs(x - median.x) > median.x/(max_deviation*10.0)) ||
					(fabs(y - median.y) > max_deviation && fabs(y - median.y) > median.y/(max_deviation*10.0)) )
			{
				// estimate the position
				num_predicted_features++;
				curr_status[i] >= FLOW_STATUS_OK ? curr_status[i]++ : curr_status[i] = FLOW_STATUS_ESTIMATED;
			}
			// otherwise it was a good point, so calculate the exact direction
			else
			{
				num_found_features++;
				avg_x += x;
				avg_y += y;
			}
		}
	}

	// XXX: maybe add a threshold, when less then 10% of features found
	//if (!num_found_features)
	if (num_found_features < 10)
	{
		CvPoint2D32f lastdir = prev->getDirection();
		fprintf(stderr, "Not enough (%d) good points found anymore. Guessing direction from last frame: %.2f,%.2f\n", num_found_features, lastdir.x, lastdir.y);
		for(int i =0; i<num_features; i++)
		{
		    features[i].x = prev->features[i].x + lastdir.x;
		    features[i].y = prev->features[i].y + lastdir.y;
			curr_status[i] = prev->status[i] >= FLOW_STATUS_ESTIMATED ?
				prev->status[i] + 1 : FLOW_STATUS_ESTIMATED;
			if(curr_status[i] >= FLOW_STATUS_LOST)
			{
				curr_status[i] = FLOW_STATUS_LOST;
				lost_features++;
			}
		}
		return lost_features;
	}

	avg_x /= num_found_features;
	avg_y /= num_found_features;

	// adjust bad marked points by average direction of good points
	for(int i=0; i < num_features; i++) 
	{
		if (curr_status[i] == FLOW_STATUS_NOT_OK || curr_status[i] >= FLOW_STATUS_ESTIMATED)
		{
		    features[i].x = prev->features[i].x + avg_x;
		    features[i].y = prev->features[i].y + avg_y;
			curr_status[i] = prev->status[i] >= FLOW_STATUS_ESTIMATED ?
				prev->status[i] + 1 : FLOW_STATUS_ESTIMATED;
			if(curr_status[i] >= FLOW_STATUS_LOST)
			{
				curr_status[i] = FLOW_STATUS_LOST;
				lost_features++;
			}
		}
	}
	return lost_features;
}

// this functions tries to calculate new coordindates
// according to the difference of previous feature points
// and stores the guessed features in the out parameter `feature'
// which has to be preallocated with enough room
void OpticalFlow::guess(CvPoint2D32f* features, TrackingObject *prev)
{
	CvPoint2D32f direction = prev->getDirection();

	for(int i=0; i < num_features; i++) 
	{
		features[i].x = curr_features[i].x + direction.x;
		features[i].y = curr_features[i].y + direction.y;
	}
}


// INPUT: a gray scale image
// OUTPUT: features, status of the features, and number of the features
//
// be sure to use/copy the returned features before the next call, because
// this function only returns a pointer to static memory and does not copy the features
int OpticalFlow::updateFlow(const IplImage* next_image, CvPoint2D32f** features = NULL, char** status = NULL, int* lost_feature_count = NULL)
{
	if (num_features == 0)
	{
		fprintf(stderr, "No features selected, unable to continue. Aborting...\n");
		exit(EXIT_FAILURE);
	}

#if 0
	printf("old status: ");
	for(int i =0; i<num_features; i++)
		printf("%d|", curr_status[i]);
	printf("\n");
#endif

	CvPoint2D32f final_features[MAX_FEATURE_POINTS];
	CvPoint2D32f temp_features[MAX_FEATURE_POINTS];
	char final_status[MAX_FEATURE_POINTS];
	char temp_status[MAX_FEATURE_POINTS];
	memset(final_status, FLOW_STATUS_ESTIMATED, sizeof(char) * num_features);

	guess(guess_features, tracking_objects->back());

	int good = 0;
	int flags = CV_LKFLOW_INITIAL_GUESSES; // clear all flags but this one
	std::vector<TrackingObject*>::const_reverse_iterator rii;
	for(rii = tracking_objects->rbegin(); rii!=tracking_objects->rend(); ++rii)
	{
		TrackingObject* last = *rii;
		IplImage* old_image = last->gray_image;
		CvPoint2D32f* old_features = last->features;
		char* old_status = last->status;

		// skip this frame, if it does not contain enough relevant features
		int temp_num = 0;
		for (int i=0; i<num_features; i++)
		{
			if(old_status[i] == FLOW_STATUS_OK && final_status[i] != FLOW_STATUS_OK)
			{
				temp_features[temp_num] = old_features[i];
				curr_features[temp_num] = guess_features[i];
				temp_num++;
			}
		}
		if(!temp_num || (temp_num < good/10))
			continue;

		// do the actual calculation of the feature points
		cvCalcOpticalFlowPyrLK( old_image, next_image, prev_pyramid, curr_pyramid,
				temp_features, curr_features, temp_num, cvSize(FLOW_WIN_SIZE, FLOW_WIN_SIZE), 3, temp_status, error,
				cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,20,0.03), flags );
		flags |= CV_LKFLOW_PYR_B_READY;

		int curr_temp_num_index = 0;
		for (int i=0; i<num_features; i++)
		{
			if(old_status[i] != FLOW_STATUS_OK || final_status[i] == FLOW_STATUS_OK)
				continue;

			if (temp_status[curr_temp_num_index] == FLOW_STATUS_OK && error[curr_temp_num_index] < FLOW_MAX_ERROR)
			{
				good++;
				final_features[i] = curr_features[curr_temp_num_index];
				final_status[i] = FLOW_STATUS_OK;
			}
			else // features position will be calculated anyway, don't need to set it here
				final_status[i] = FLOW_STATUS_ESTIMATED;

			curr_temp_num_index++;
			if(curr_temp_num_index > temp_num)
				fprintf(stderr, "ASSERT: curr_temp_num_index > temp_num\n");
		}
	}
	
	// post processing
	memcpy(curr_status, final_status, sizeof(char) * num_features);
	int lost = correct(final_features, tracking_objects->back());
	memcpy(curr_features, final_features, sizeof(CvPoint2D32f) * num_features);

#if 0
	printf("new status: ");
	for(int i =0; i<num_features; i++)
		printf("%d|", curr_status[i]);
	printf("\n");
#endif

	if (features)
		*features = curr_features;
	if (status)
		*status = curr_status;
	if (lost_feature_count)
		*lost_feature_count = lost;

	return num_features;
}
