/* Object Tracker with high-level segementation
 *
 * (c) 2006-2007 Martin Stubenschrott
*/

// stdlib includes
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <vector>
#include <set>

#include <libgen.h> // for basename()

// opencv includes
#include "cv.h"
#include "highgui.h"

// our own includes
#include "segmentation.h"
#include "flow.h"
#include "trackingobject.h"
#include "evaluation.h"

#include "config.h"

#define DRAW_ALL_SEGMENTS     0x01
#define DRAW_ACTIVE_SEGMENTS  0x02
#define DRAW_FEATURE_POINTS   0x04
#define DRAW_BOUNDING_BOX     0x08
#define DRAW_GLOBAL_DIRECTION 0x10
#define DRAW_ALL   ((DRAW_GLOBAL_DIRECTION << 1) - 1)

// global variables
ImageSegmenter* segmenter = NULL;
OpticalFlow* flow = NULL;
PerformanceEvaluator* evaluator = NULL;
std::vector<TrackingObject*> *tracking_objects = NULL;
CvVideoWriter* video_writer = NULL;

// forward declarations
void drawFrame(IplImage *frame, ImageSegmenter *segmenter, OpticalFlow *flow, int flags);

void usage()
{
    printf ("Usage: tracker <inputvideo> [groundtruth.<txt|xml>] [skip_frames]\n\n"
            "Click in a segment of the first frame to add it to the tracked segments.\n"
			"When ready, press any key to start tracking\n\n"
			"Keys:\n"
			"\t's' writes the current image to <video_filename>_<frameno>.png\n"
			"\t'q' quit\n\n"
			"Colors:\n"
			"\tGreen: tracked feature points, tracked bounding box\n"
			"\tBlue:  estimated feature points\n"
			"\tRed:   lost feature points, recalculated when >25%% are lost; ground truth BBox\n"
			"\tPink/orange: segmented object\n");

}

void on_mouse( int event, int x, int y, int flags, void* param )
{
	IplImage* frame = (IplImage*)param;
    if( !frame )
        return;

    if( frame->origin )
        y = frame->height - y;

    if( event == CV_EVENT_LBUTTONDOWN )
    {
		if (tracking_objects->size() != 1)
		{
			fprintf(stderr, "Warning: Selecting objects is only possible on the first frame.\n");
			return;
		}

        CvPoint pt = cvPoint(x,y);
		TrackingObject* first = tracking_objects->front();
		first->addSegment(segmenter->findParentComponent(pt));

		if(evaluator)
			evaluator->point_of_id = pt;

		drawFrame(frame, segmenter, flow, DRAW_ACTIVE_SEGMENTS | DRAW_FEATURE_POINTS);
    }
}

// make sure you free the image, which was created by this function
IplImage* convertToSmoothedFloatImage(IplImage* frame)
{
	IplImage *tmp = cvCreateImage(cvGetSize(frame), IPL_DEPTH_32F, frame->nChannels);
	IplImage *frame_smoothed = cvCreateImage(cvGetSize(frame), IPL_DEPTH_32F, frame->nChannels);

	int step       = tmp->widthStep/sizeof(float);
	float *data    = (float *)tmp->imageData;
	for (int y = 0; y < tmp->height; y++)
	{
		for (int x = 0; x < tmp->width; x++)
		{
			uchar *pixel_rgb = (uchar*)(frame->imageData + y*frame->widthStep + x*3);
			float *pixel_float = (float*)(data + y*step + x*3);

			// we need to swap r,g,b because of byte ordering
			*pixel_float = *(pixel_rgb+2);
			*(pixel_float+1) = *(pixel_rgb+1);
			*(pixel_float+2) = *pixel_rgb;
		}
	}

	cvSmooth(tmp, frame_smoothed, CV_GAUSSIAN, 5, 5, GAUSSIAN_BLUR);
	cvReleaseImage(&tmp);
	return frame_smoothed;
}

// "damages" the image, be sure to copy it first, if you need the original frame afterwards
void drawFrame(IplImage *frame, ImageSegmenter *segmenter, OpticalFlow *flow, int flags = 0)
{
	if(!frame)
	{
		fprintf(stderr, "NULL pointer for frame in drawFrame()\n");
		return;
	}
	TrackingObject *curr_object = NULL;
	if (tracking_objects->size() >= 1)
		curr_object = tracking_objects->back();

#if 1
	// draw segments
	if(segmenter)
	{
		std::vector<Component*> *segments = NULL;
		if (flags & DRAW_ALL_SEGMENTS)
		{
			if(!segments)
				segments = segmenter->getSegments();

			for (unsigned int i=0; i<segments->size(); i++)
				segments->at(i)->overlayOverImage(frame, rand()%255, rand()%255, rand()%255, true);
		}
	}
	// draw active segments
	if (flags & DRAW_ACTIVE_SEGMENTS)
	{
		std::set<Component*> *active_segments = curr_object->active_segments;
		Component *comp = NULL;
		std::set<Component*>::iterator it = active_segments->begin();
		for(; it != active_segments->end(); it++)
		{
			comp = *it;
			comp->overlayOverImage(frame, 255,128,0, false);
		}
	}

	// draw current feature points
	if(flags & DRAW_FEATURE_POINTS)
	{
		const int step = 127 / (FLOW_STATUS_LOST - FLOW_STATUS_OK);
		for (int i=0; i<curr_object->num_features; i++)
		{
			if (curr_object->status[i] == FLOW_STATUS_OK) //feature point found
				cvCircle(frame, cvPointFrom32f(curr_object->features[i]), 1, cvScalar(0,255,0), 1);
			else if (curr_object->status[i] >= FLOW_STATUS_ESTIMATED && curr_object->status[i] < FLOW_STATUS_LOST) // manually said that not found
			{
				int color = 128 + (curr_object->status[i] - FLOW_STATUS_ESTIMATED + 1) * step;
				cvCircle(frame, cvPointFrom32f(curr_object->features[i]), 1, cvScalar(color,128,0), 1);
			}
			else // not found at all -> lost
				cvCircle(frame, cvPointFrom32f(curr_object->features[i]), 1, cvScalar(0,0,255), 1);
		}
	}

	// draw direction arrow
	if(curr_object && (flags & DRAW_GLOBAL_DIRECTION))
	{
		CvPoint center = cvPointFrom32f(curr_object->getCenter());
		CvPoint last_dir = cvPointFrom32f(curr_object->getDirection());

		// center point
		cvCircle(frame, center, 4, cvScalar(255,255,255), 1, CV_AA);
		// last direction
		cvLine(frame, center, cvPoint(center.x + last_dir.x, center.y + last_dir.y),  cvScalar(0,255,255), 2, CV_AA);
	}

	if(flags & DRAW_BOUNDING_BOX)
	{
		CvRect bbox;
		// in red the correct bbox
		if(evaluator)
		{
			bbox = evaluator->curr_gt_bbox;
			cvRectangle(frame, cvPoint(bbox.x, bbox.y), cvPoint(bbox.x + bbox.width, bbox.y + bbox.height), CV_RGB(255,0,0), 1, CV_AA);
			cvCircle(frame, cvPoint(evaluator->curr_gt_center.x, evaluator->curr_gt_center.y), 4, CV_RGB(255,0,0), 1, CV_AA | CV_FILLED);
		}
		bbox = curr_object->getBoundingBox();
		// in green our calculated one
		cvRectangle(frame, cvPoint(bbox.x, bbox.y), cvPoint(bbox.x + bbox.width, bbox.y + bbox.height), CV_RGB(0,255,0), 1, CV_AA);
	}
#endif
	cvShowImage(WINDOW_NAME, frame);

	// also write to video output
	if(video_writer)
	{
		// show the first frame longer in the video
		int times = 1;
		if (tracking_objects->size() == 1)
			times = 24 * 5; // 5 seconds
		for (int i=0; i < times; i++)
			cvWriteFrame(video_writer, frame);
	}
}

/* how the main() function works:
 *
 * segment_and_show_first_frame();
 * user selects all segments with the mouse which belong to the object
 * for each selected segment, extract good feature points to track
 *
 * for all remaining frames:
 *   calculate_optflow(featurepoints);
 *   check_and_restore_missing_featurepoints();
 *   segment_frame();
 *   segments where tracked feature points lie in are labelled as part of the object
 *
*/
int main( int argc, char** argv )
{
	long skip_frames = 0; // if the video does not start with the first frame
	char* ground_truth_file = NULL;

	// check our arguments or show help message
	if (argc < 2 || argc > 4 || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--help"))
	{
		usage();
		exit(EXIT_FAILURE);
	}
	else if (argc == 3)
	{
		errno=0;
		char *endptr;
		long val = strtol(argv[2], &endptr, 10);
		if(!errno && (*endptr == '\n' || *endptr == '\0'))
			skip_frames = val;
		else // no conversion done, use this argument as GT
			ground_truth_file = argv[2];
	}
	else
	{
		ground_truth_file = argv[2];
		skip_frames = strtol(argv[3], (char **)NULL, 10);
	}

	// our inputvideo
    CvCapture* capture = cvCaptureFromAVI( argv[1] );
    if(!capture)
    {
        fprintf(stderr, "Could not initialize capturing...\n");
        return EXIT_FAILURE;
    }

	// process first frame individually to get a region of interest
	IplImage *frame = NULL, *frame_smoothed = NULL, *frame_segmentation = NULL, *frame_gray = NULL;

	// initialization of the evaluator
	if (ground_truth_file != NULL)
	{
		// Note that this detection scheme is _very_ simple and error prone,
		// so please don't try to fool it or provide a bogus evaluation data file
		char *pos = strcasestr(ground_truth_file, ".txt");
		if(pos != NULL)
			evaluator = new MusclePerformanceEvaluator (ground_truth_file, skip_frames+1);
		else
			evaluator = new CaviarPerformanceEvaluator (ground_truth_file, skip_frames+1);
	}

	// skip first n frames, if specified
	for (int i=0;i<skip_frames; i++)
		frame = cvQueryFrame( capture );

	frame = cvQueryFrame( capture );
	frame_gray = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
	cvCvtColor(frame, frame_gray, CV_RGB2GRAY);
	skip_frames++; // keep track how many frames were already skipped/processed
	if (!frame || !frame_gray)
    {
        fprintf(stderr, "Could not read first frame...\n");
        return EXIT_FAILURE;
    }

	// our outputvideo
#ifdef WRITE_VIDEO
    video_writer = cvCreateVideoWriter("output.avi", CV_FOURCC('x','v','i','d'), 25,
			cvSize(frame->width, frame->height));
	if (!video_writer)
		fprintf(stderr, "Warning: cvCreateVideoWriter failed, output.avi will not be created!\n");

	// show the original first frame
	int times = 24 * 5; // 4 seconds
	for (int i=0; i < times; i++)
		cvWriteFrame(video_writer, frame);

#endif

	// initialization of the performance evaluator and optical flow class
	flow = new OpticalFlow(frame_gray);

	// initialization of the segmenter
	segmenter = new ImageSegmenter();
	frame_smoothed = convertToSmoothedFloatImage(frame);
	segmenter->setImage(frame_smoothed);
	std::vector<Component*> *all_segments = segmenter->getSegments();

	// initialization of the object tracker
	tracking_objects = new std::vector<TrackingObject*>;
	TrackingObject *first_object = new TrackingObject(frame_gray);
	tracking_objects->push_back(first_object);
	
	// we have our first frame, initialize our window and a mouse handler
	frame_segmentation = cvCloneImage(frame);
    cvNamedWindow (WINDOW_NAME, 0);
    cvSetMouseCallback (WINDOW_NAME, on_mouse, frame_segmentation );

	drawFrame(frame, segmenter, flow, DRAW_ALL_SEGMENTS);

	if ((char)cvWaitKey(-1) == 'q') // wait endlessly until some key is pressed
		goto CLEANUP;

	cvReleaseImage(&frame_smoothed);
	cvReleaseImage(&frame_segmentation);

	// for our segments, try to find some good feature points
	first_object->addFeaturePoints();
	// if no feature points were selected, exit
	if (first_object->num_features <= 0)
	{
		fprintf(stderr, "Error: no feature points in main()!\n");
		goto CLEANUP;
	}
	first_object->finalize(NULL);

	// now we have all our initial features, add them to the optical flow object
	flow->setCurrentFeaturePoints(first_object->features, first_object->status, first_object->num_features);


	// main loop to scan each remaining frame of the video
	while(1)
    {
        frame = cvQueryFrame( capture );
        if( !frame )
            break;

		printf("============ next frame: %ld =================\n", skip_frames++);
		//if (skip_frames > 20) break;

		// convert to gray image for optical flow calculation
		IplImage *gray_frame = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
		cvCvtColor(frame, gray_frame, CV_RGB2GRAY);

		// calculate new feature point positions
		CvPoint2D32f* features;
		char* status;
		int lost_count;
		int count = flow->updateFlow(gray_frame, &features, &status, &lost_count);

		// segment the new image and extract segments where feature points are located
		frame_smoothed = convertToSmoothedFloatImage(frame);
		segmenter->setImage(frame_smoothed);
		//std::vector<Component*> *all_segments = segmenter->getSegments(); // calculate new segmentation
		all_segments = segmenter->getSegments(); // calculate new segmentation
		std::set<Component*> *active_segments = segmenter->getActiveSegments(features,status,count);

		// creating new TrackingObject
		TrackingObject* object = new TrackingObject(gray_frame, features, status, count);
		std::set<Component*>::iterator it;
		for(it = active_segments->begin(); it != active_segments->end(); it++)
			object->addSegment(*it);

		//segmenter->setFeaturePoints(features, status, count);
		// find lost feature points
		if(lost_count > count/4)
		{
			fprintf(stderr, "Warning: More than 25%% of features (%d of %d) lost: Regenerating...\n", lost_count, count);
			object->addFeaturePoints();
		}
		cvReleaseImage(&frame_smoothed);

		object->finalize(tracking_objects->back());
		
		// add it to our vector of previous tracking objects
		if(evaluator)
			evaluator->addFrameInformation(object);
		tracking_objects->push_back(object);
		if (tracking_objects->size() > MAX_TRACKING_OBJECTS)
		{
			// NOTE: erasing at the beginning has O(n), at the end O(1),
			// but for our cases this doesn't really matter, maybe sometime switch to std::list or std::dequeue
			std::vector<TrackingObject*>::iterator it = tracking_objects->begin();
			TrackingObject *to = *it;
			delete to;
			tracking_objects->erase(it);
		}

		//drawFrame(frame, segmenter, flow, DRAW_ACTIVE_SEGMENTS | DRAW_FEATURE_POINTS | DRAW_GLOBAL_DIRECTION);
		drawFrame(frame, segmenter, flow, DRAW_ACTIVE_SEGMENTS | DRAW_FEATURE_POINTS | DRAW_GLOBAL_DIRECTION | DRAW_BOUNDING_BOX);
		//drawFrame(frame, segmenter, flow, DRAW_ACTIVE_SEGMENTS);
		char key = (char)cvWaitKey(FRAME_WAIT_TIME);
		if (key == 's') // save image
		{
			char* name = basename(argv[1]);
			char fullname[PATH_MAX] = "";
			if(!name)
				break;
			snprintf(fullname, PATH_MAX-1, "%s_%ld.png", name, skip_frames);
			fullname[PATH_MAX-1] = '\0'; // make sure we nullterminate the string to don't allow buffer overflows

			if(!cvSaveImage(fullname, frame))
				fprintf(stderr, "Warning: Could not save: %s\n", fullname);
		}
		else if (key == 'q') // must be called after cvShowImage, to really show the image
			break;
	}

	// Finally print some tracking evaluation information
	if(evaluator)
	{
		printf("========= Performance Report: ==============\n");
		printf( "ID: %d\n"
				"Successfully tracked: %3.2f%%\n"
				"Overlap:              %3.2f%%\n",
				evaluator->getID(),
				evaluator->successfullyTracked()*100,
				evaluator->overlappingBoundingBoxes()*100);
	}
	
	// clean up
CLEANUP:
	if(video_writer)
		cvReleaseVideoWriter( &video_writer );

    cvReleaseCapture (&capture);
    cvDestroyWindow (WINDOW_NAME);
	delete segmenter;
	delete flow;
	delete evaluator;
	for( unsigned int i=0; i < tracking_objects->size() ; i++ )
		delete tracking_objects->at(i);
	tracking_objects->clear();
	delete tracking_objects;

    return EXIT_SUCCESS;
}
