#include "cv.h"
#include "highgui.h"
#include "segmentation.h"
#include "config.h"

#include <vector>
#include <map>
#include <algorithm>
#include <iostream>
#include <stdlib.h>
#include <math.h>

inline bool Component::addPoint(CvPoint pt)
{
	bool ret = false;
	if (!points)
	{
		points = new std::vector<CvPoint>();
		min_x = max_x = pt.x;
		min_y = max_y = pt.y;
		ret = true;
	}
	points->push_back(pt);
	// adjust bounding box
	min_x = MIN(min_x, pt.x);
	min_y = MIN(min_y, pt.y);
	max_x = MAX(max_x, pt.x);
	max_y = MAX(max_y, pt.y);

	return ret;
}
void Component::overlayOverImage(IplImage* img, uchar red, uchar green, uchar blue, bool transparent)
{
	if (!points)
	{
		fprintf(stderr, "points == NULL in overlayOverImage()\n");
		return;
	}

	for (unsigned int i = 0; i < points->size(); i++)
	{
		//float *pixel = (float*)(img->imageData + points->at(i).y*img->widthStep + points->at(i).x*img->nChannels * sizeof(float));
		uchar *pixel = (uchar*)(img->imageData + points->at(i).y*img->widthStep + points->at(i).x*img->nChannels * sizeof(uchar));
		if (blue > 0)
			transparent ? *pixel = (*pixel+blue) / 2 : *pixel = blue;
		if (green > 0)
			transparent ? *(pixel+1) = (*(pixel+1)+green) / 2 : *(pixel+1) = green;
		if (red > 0)
			transparent ? *(pixel+2) = (*(pixel+2)+red) / 2 : *(pixel+2) = red;
	}

	// bounding box
	//cvRectangle(img, cvPoint(min_x, min_y), cvPoint(max_x, max_y), CV_RGB(255,0,0), 1, 4, 0);
}

// stores found feature points in preallocated features array
// uses the status array to determine if we need to find a new feature point
//
// max_count determines how many points should be ADDED at maximum, not how
// many can be in the features array at the end
//
// returns: total number of good features in features array
int Component::findGoodFeaturePoints(IplImage *gray_img, CvPoint2D32f* features, char* status, int max_count)
{
	if(!points || !gray_img || !features || !status || !max_count)
	{
		fprintf(stderr, "NULL pointer arguments in findGoodFeaturePoints... %p, %p, %p\n", points, gray_img, features);
		return 0;
	}

	IplImage* eig = cvCreateImage( cvGetSize(gray_img), 32, 1 );
	IplImage* temp = cvCreateImage( cvGetSize(gray_img), 32, 1 );
	IplImage* mask = cvCreateImage( cvGetSize(gray_img), IPL_DEPTH_8U, 1 );
	IplImage* mask_eroded = cvCreateImage( cvGetSize(gray_img), IPL_DEPTH_8U, 1 );
	IplConvKernel *se = cvCreateStructuringElementEx(3, 3, 1, 1, CV_SHAPE_CROSS);

	int total_count = 0, good_count = 0, added_count = 0;
	CvPoint2D32f good_features[MAX_FEATURE_POINTS];
   	
   	// set mask
   	cvZero(mask);
   	for (unsigned int i = 0; i < points->size(); i++)
   	{
   		uchar *pixel = (uchar*)(mask->imageData + points->at(i).y*mask->widthStep + points->at(i).x*mask->nChannels * sizeof(uchar));
   		*pixel = 255;
   	}
   	cvErode(mask, mask_eroded, se, 3);
   	
#if 0
	cvNamedWindow ("mask", 0);
	cvNamedWindow ("mask_eroded", 0);
	cvShowImage( "mask", mask );
	cvShowImage( "mask_eroded", mask_eroded );
	cvWaitKey(10000);
#endif

	double quality = 0.0001;
	double min_distance = 3;
	good_count = MAX_FEATURE_POINTS;
	cvGoodFeaturesToTrack( gray_img, eig, temp, good_features, &good_count,
			quality, min_distance, mask_eroded, 3, 0, 0.04 );

	// XXX: maybe improves our features
	//	cvFindCornerSubPix( gray_img, points[1] + count - 1, 1,
	//			cvSize(win_size,win_size), cvSize(-1,-1),
	//			cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,20,0.03));


	// -> try to add some extra features
	
	CvPoint r;
	CvPoint2D32f pt;
	int consecutive_errors=0;

	// only generate random points in the bounding box
	int width = max_x - min_x;
	int height = max_y - min_y;
	if(!width || !height)
		goto CLEANUP;

	for(int i = 0; i < MAX_FEATURE_POINTS; i++)
	{
		if(status[i] == FLOW_STATUS_OK)
		{
			total_count++;
			continue;
		}
		else if (status[i] != FLOW_STATUS_LOST) // don't rebuild estimated points
			continue;

		if (added_count >= max_count)
			break;

		while(consecutive_errors < 50)
		{

			// first add good features, then random
			if(good_count)
				pt = good_features[--good_count];
			else // random
			{
				pt.x = min_x + rand() % width;
				pt.y = min_y + rand() % height;
			}
			r = cvPointFrom32f(pt);

			// TODO: obey minimum distance
			uchar *pixel = (uchar*)(mask_eroded->imageData + r.y*mask_eroded->widthStep + r.x*mask_eroded->nChannels * sizeof(uchar));
			if(*pixel) // pixel is in current mask
			{
				consecutive_errors = 0;
				added_count++;
				total_count++;
				status[i] = FLOW_STATUS_OK;
				features[i] = pt;
				break;
			}
			else
			{
				consecutive_errors++;
				continue;
			}
		}
	}

CLEANUP:
	printf("Features added: total: %d, added: %d (%d good)\n", total_count, added_count, good_count);

	cvReleaseImage( &eig );
	cvReleaseImage( &temp );
	cvReleaseImage( &mask );
	cvReleaseImage( &mask_eroded );
	cvReleaseStructuringElement(&se);
	
	return total_count;
}
/*
 * CONSTRUCTOR
 */
ImageSegmenter::ImageSegmenter()
{
    mem_storage = cvCreateMemStorage(0);
	image = NULL;
    graph = NULL; // graph is built, when the first image is set with setImage()
	edges = NULL;
	segments = NULL;
//	active_segments = NULL;
//	num_features = 0;
}

/*
 * DESTRUCTOR
 */
ImageSegmenter::~ImageSegmenter()
{
	freeResources();
}

void ImageSegmenter::freeResources()
{
    // clear the graph first, if it was already built
    if(graph != NULL)
        cvClearGraph(graph);
	graph = NULL;

	if (edges)
	{
		free(edges);
		edges = NULL;
	}

//	if (active_segments)
//	{
//		active_segments->clear();
//		delete active_segments;
//		active_segments = NULL;
//	}

	if (segments)
	{
		for(unsigned int i=0; i<segments->size(); i++)
		{
			// the rest of the Component* was already cleard by cvClearGraph above
			segments->at(i)->free();
		}
		segments->clear();
		delete segments;
		segments = NULL;
	}

	if (mem_storage)
		cvClearMemStorage(mem_storage);
}

void ImageSegmenter::setImage(IplImage* img)
{
	freeResources();
	image = img;
    initGraph(8);
}

/* input: color image
 *
 * output: segmented regions as sequences of points, don't forget
 *         to delete the vector
 */
std::vector<Component*>* ImageSegmenter::getSegments()
{
    if(segments) //segments already calculated, just need to return them
		return segments;

	segments = new std::vector<Component*>;

    if (!image)
    {
        std::cerr << "image == NULL, in: getSegments()" << std::endl;
        delete segments;
        return NULL;
    }

    // 1. sort edges in non-descending order
	// std::sort(weights.begin(), weights.end(), AscendingWeightSort());
    std::sort(edges, edges+num_edges, AscendingWeightSort());

	// 2. merge points which edges have little weight
//    for(int i=0;i<weights.size();i++)
    for(int i=0;i<num_edges;i++)
    {
//		printf("weight: %f\n", edges[i]->weight);
//        Component* v1 = findParentComponent((Component*)weights.at(i)->vtx[0]);
//        Component* v2 = findParentComponent((Component*)weights.at(i)->vtx[1]);
        Component* v1 = findParentComponent((Component*)edges[i]->vtx[0]);
        Component* v2 = findParentComponent((Component*)edges[i]->vtx[1]);

		// checks are all done in the function
		//mergeComponents(v1, v2, weights.at(i));
		mergeComponents(v1, v2, edges[i]);
    }

	int num_before_merge = num_segments;
	// 3. merge too small components
	//for(int i=0;i<weights.size();i++)
	for(int i=0;i<num_edges;i++)
	{
        Component* v1 = findParentComponent((Component*)edges[i]->vtx[0]);
        Component* v2 = findParentComponent((Component*)edges[i]->vtx[1]);
		
		if (v1 != v2 && (v1->size < SEGMENTATION_MIN_COMPONENT_SIZE || v2->size < SEGMENTATION_MIN_COMPONENT_SIZE))
			mergeComponents(v1, v2, edges[i], true); // a weight of 0 means, always merge
	}
	printf("Number of total segments: %d (was %d before merging small components)\n", num_segments, num_before_merge);

	// 4. store the pixels for later drawing
	int index=0;
	for (int y = 0; y < image->height; y++)
	{
		for (int x = 0; x < image->width; x++,index++)
		{
			Component* pt = (Component*)cvGetGraphVtx(graph, index);
			Component* comp = findParentComponent(pt);

			// add this point to the component and
			if(comp->addPoint(cvPoint(x,y)))
				segments->push_back(comp);
		}
	}

    return segments;
}

// also calculates weights between nodes
void ImageSegmenter::initGraph(int connectivity)
{
    graph = cvCreateGraph(CV_SEQ_KIND_GRAPH, sizeof(CvGraph), sizeof(Component), sizeof(CvGraphEdge), mem_storage);
	edges = (CvGraphEdge**) malloc( sizeof(CvGraphEdge*) * image->height * image->width * 4);

    CvGraphEdge *edge = NULL;
    Component* vertex = NULL;
	num_edges = 0;
    for (int y=0; y < image->height; y++)
    {
        for (int x=0; x < image->width; x++)
        {
            //printf("Adding vertex index: %d - %d - %d - width: %d - widthStep:%d\n", cvGraphAddVtx(graph), x, y, image->width, image->widthStep);
            cvGraphAddVtx(graph, NULL, (CvGraphVtx**)&vertex);
			vertex->init();

            // add 4 edges (to the left, to the top, to the top-left an to the top-right)
            // for each node which is not at the top or left or right border
            if (x > 0 && y > 0 && x < image->width-1)
			{
				cvGraphAddEdge(graph, y * image->width+x, y * image->width+x-1, NULL, &edge);     // left
				edge->weight = getWeight(cvPoint(x,y), cvPoint(x-1,y));
				//weights.push_back(edge);
				edges[num_edges++] = edge;
				cvGraphAddEdge(graph, y * image->width+x, (y-1) * image->width+x, NULL, &edge);   // top
				edge->weight = getWeight(cvPoint(x,y), cvPoint(x,y-1));
				//weights.push_back(edge);
				edges[num_edges++] = edge;
				if(connectivity == 8)
				{
					cvGraphAddEdge(graph, y * image->width+x, (y-1) * image->width +x-1, NULL, &edge); // topleft
					edge->weight = getWeight(cvPoint(x,y), cvPoint(x-1,y-1));
					//weights.push_back(edge);
					edges[num_edges++] = edge;
					cvGraphAddEdge(graph, y * image->width+x, (y-1) * image->width +x+1, NULL, &edge); // topright
					edge->weight = getWeight(cvPoint(x,y), cvPoint(x+1,y-1));
					//weights.push_back(edge);
					edges[num_edges++] = edge;
				}
			}
        }
    }
    // post process to add edges for the x==0 and y==0
    for (int x=1; x < image->width; x++) // first row
    {
        cvGraphAddEdge(graph, x, x-1, NULL, &edge); //left
        edge->weight = getWeight(cvPoint(x,0), cvPoint(x-1,0));
//		weights.push_back(edge);
    }
    for (int y=1; y < image->height; y++) 
    {
        // first column
        cvGraphAddEdge(graph, y * image->width, (y-1) * image->width, NULL, &edge); //top
        edge->weight = getWeight(cvPoint(0,y-1), cvPoint(0,y));
//		weights.push_back(edge);
		if(connectivity == 8)
		{
			cvGraphAddEdge(graph, y * image->width, (y-1) * image->width + 1, NULL, &edge); //topright
			edge->weight = getWeight(cvPoint(1,y-1), cvPoint(0,y));
//			weights.push_back(edge);
		}

        // last column
        cvGraphAddEdge(graph, ((y+1) * image->width)-1, y * image->width -1, NULL, &edge); //top
        edge->weight = getWeight(cvPoint(image->width-1,y-1), cvPoint(image->width-1,y));
//		weights.push_back(edge);
        cvGraphAddEdge(graph, ((y+1) * image->width)-1, ((y+1) * image->width)-2, NULL, &edge); //left
        edge->weight = getWeight(cvPoint(image->width-1,y), cvPoint(image->width-2,y));
//		weights.push_back(edge);
		if(connectivity == 8)
		{
			cvGraphAddEdge(graph, ((y+1) * image->width)-1, (y * image->width)-2, NULL, &edge); //topleft
			edge->weight = getWeight(cvPoint(image->width-2,y-1), cvPoint(image->width-1,y));
//			weights.push_back(edge);
		}
    }

	num_segments = image->height * image->width;
}

inline float ImageSegmenter::getWeight(CvPoint pt1, CvPoint pt2)
{
    //uchar blue1   = *(image->imageData + pt1.y*image->widthStep + pt1.x*3*sizeof(float));
    //uchar green1 = *(image->imageData + pt1.y*image->widthStep + pt1.x*3*sizeof(float)+1);
    //uchar red1  = *(image->imageData + pt1.y*image->widthStep + pt1.x*3*sizeof(float)+2);

    //uchar blue2   = *(image->imageData + pt2.y*image->widthStep + pt2.x*3*sizeof(float));
    //uchar green2 = *(image->imageData + pt2.y*image->widthStep + pt2.x*3*sizeof(float)+1);
    //uchar red2  = *(image->imageData + pt2.y*image->widthStep + pt2.x*3*sizeof(float)+2);
	
    float blue1   = *(float*)(image->imageData + pt1.y*image->widthStep + pt1.x*3*sizeof(float));
    float green1 = *(float*)(image->imageData + pt1.y*image->widthStep + pt1.x*3*sizeof(float)+1*sizeof(float));
    float red1  = *(float*)(image->imageData + pt1.y*image->widthStep + pt1.x*3*sizeof(float)+2*sizeof(float));

    float blue2   = *(float*)(image->imageData + pt2.y*image->widthStep + pt2.x*3*sizeof(float));
    float green2 = *(float*)(image->imageData + pt2.y*image->widthStep + pt2.x*3*sizeof(float)+1*sizeof(float));
    float red2  = *(float*)(image->imageData + pt2.y*image->widthStep + pt2.x*3*sizeof(float)+2*sizeof(float));

    float diff = sqrt(pow(blue1-blue2, 2.f) + pow(green1-green2, 2.f) + pow(red1-red2, 2.f));

    return diff;
}

inline float ImageSegmenter::minimalInternalDifference(Component *vtx1, Component *vtx2)
{
	float diff1 = vtx1->internalDifference + (SEGMENTATION_THRESHOLD / vtx1->size);
	float diff2 = vtx2->internalDifference + (SEGMENTATION_THRESHOLD / vtx2->size);

    return diff1 < diff2 ? diff1 : diff2;
}

inline bool ImageSegmenter::mergeComponents(Component *vtx1, Component *vtx2, CvGraphEdge* edge, bool force)
{
	if (vtx1 == vtx2)   // never merge identical components
		return false;
	if (!vtx1 || !vtx2 || !edge) // null pointer check
		return false;

//	printf("index: %d,%d, int: %f,%f, edge: %f, size: %d,%d, rank: %d, %d\n",cvGraphVtxIdx(graph,edge->vtx[0]), cvGraphVtxIdx(graph,edge->vtx[1]), vtx1->internalDifference, vtx2->internalDifference, edge->weight, vtx1->size, vtx2->size, vtx1->rank, vtx2->rank);

	if (!force && edge->weight > minimalInternalDifference(vtx1, vtx2))
		return false;

//	printf("Merge!\n");
	if (vtx1->rank > vtx2->rank)
	{
		vtx2->parent = vtx1;
		vtx1->size += vtx2->size;
		
		// update internal difference to this edge, if this edge's weight is larger
		vtx1->internalDifference = edge->weight;
	}
	else
	{
		vtx1->parent = vtx2;
		vtx2->size += vtx1->size;
		if (vtx1->rank == vtx2->rank)
			vtx2->rank++;
		
		// update internal difference to this edge, if this edge's weight is larger
		vtx2->internalDifference = edge->weight;

	}
	num_segments--;
	return true;
}

Component* ImageSegmenter::findParentComponent(CvPoint pt)
{

	Component* vtx = (Component*)cvGetGraphVtx(graph, pt.y * image->width + pt.x);

	return findParentComponent(vtx);
}

inline Component* ImageSegmenter::findParentComponent(Component *child)
{
	if (!child)
	{
		fprintf(stderr, "Error: cannot find parent for NULL pointer child in findParentComponent\n");
		return NULL;
	}

	Component *res = child;
	while(res->parent != res)
		res = res->parent;

	//and then store the shorter path for future access (path compression)
	child->parent = res;
	return res;
}

// returns static memory. Don't free() it and copy result when the result needs to be kept longer
std::set<Component*>* ImageSegmenter::getActiveSegments(CvPoint2D32f* features, char* status, int num_features)
{
	static std::set<Component*> active_segments;
	std::map<Component*,int> counted_segments;

	if (!segments)
	{
		fprintf(stderr, "Warning: segments == NULL in findActiveSegments(), calling getSegments()\n");
		getSegments();
	}

	for (int i=0; i<num_features; i++)
	{
		if( status[i] == FLOW_STATUS_OK )
		{
			CvPoint pt = cvPointFrom32f(features[i]);
			Component* parent = findParentComponent(pt);
			if (!parent) // happens when a feature point is offscreen
			{
				fprintf(stderr, "Warning: feature point off-screen: %f, %f\n", features[i].x, features[i].y);
				continue;
			}
			std::map<Component*,int>::iterator it = counted_segments.find(parent);
			if (it == counted_segments.end())
				counted_segments[parent] = 1;
			else
			{
				int count = (*it).second;
				counted_segments[parent] = count+1;
			}
		}
	}

	// only leave these segments which consists of at least 3 feature points
	active_segments.clear();
	for(std::map<Component*,int>::const_iterator it = counted_segments.begin();
			it != counted_segments.end(); it++)
	{
		int count = (*it).second;
		if(count >= 3)
			active_segments.insert((*it).first);
	}
	return &active_segments;
}
