#ifndef __SEGMENTATION_H__
#define __SEGMENTATION_H__

#include "cv.h"
#include "config.h"
#include <vector>
#include <set>

// This struct is used by std::sort to sort our edges in non-decreasing order
struct AscendingWeightSort
{
	inline bool operator()(const CvGraphEdge* edge1, const CvGraphEdge* edge2)
	{
		return edge1->weight < edge2->weight;
	}
};

// this modified CvGraphVtx struct is our structure for storing
// components. It therefore adds a field of other vertices
// which belong to this compontent
//class Component
struct Component
{
    CV_GRAPH_VERTEX_FIELDS()

	Component *parent;

	float internalDifference;
	int rank;
	int size;

	std::vector<CvPoint> *points;
	// bounding box
	int min_x, min_y, max_x, max_y;

	// return true if this is the first point of the segment
	bool addPoint(CvPoint pt);
	// draws this object on top of an image with a green color and bounding box
	void overlayOverImage(IplImage* img, uchar red, uchar green, uchar blue, bool transparent);
	// finds points with biggest eigen values as good feature points, and returns the count
	int findGoodFeaturePoints(IplImage *gray_img, CvPoint2D32f* features, char* status, int max_count);
	
	// this init() function must be called because cvGraphAddVtx does not call our constructor
	inline void init() {
		parent = this;
		rank = 0;
		size = 1;
		internalDifference = 0.f; //threshold;
		points = NULL;
	}
	inline void free() {
		points->clear();
		delete points;
	}
};

/* The ImageSegmenter class reads an input image and returns a sequence 
 * of segments
 *
 * This implementation uses Felzenszwalb+Huttenlocher algorithm
 */
class ImageSegmenter
{
	public:
		ImageSegmenter(); // empty constructor, set image with setImage()
		virtual ~ImageSegmenter();

		void setImage(IplImage* img);
		//void setFeaturePoints(CvPoint2D32f* features, char* status, int num_features);

		// returns a sequence of all Segment objects
		virtual std::vector<Component*>* getSegments();
		// returns a sequence of all currentyl active segments, which have feature points in it
		std::set<Component*>* getActiveSegments(CvPoint2D32f* features, char* status, int num_features);

		Component* findParentComponent(CvPoint pt);
	
	private:
		void initGraph(int connectivity);
		void freeResources();
		float getWeight(CvPoint pt1, CvPoint pt2);
		float minimalInternalDifference(Component *vtx1, Component *vtx2);
		bool mergeComponents(Component *vtx1, Component *vtx2, CvGraphEdge* edge, bool force=false);
		Component* findParentComponent(Component *child);

		IplImage *image;
		CvGraph *graph;
		CvMemStorage* mem_storage;
		//std::vector<CvGraphEdge*> weights; // sorted sequence of edges
		CvGraphEdge **edges; // sorted sequence of edges
		int num_segments, num_edges;//, num_features;

		// this field is std::vector, because we save time for not checking set
		// properties when inserting elements
		std::vector<Component*> *segments; 
		//std::set<Component*> *active_segments;

		//CvPoint2D32f features[MAX_FEATURE_POINTS];
		//char feature_status[MAX_FEATURE_POINTS];
};


#endif /* __SEGMENTATION_H__ */
