#include <math.h>
#include <limits.h>
#include <vector>
#include <algorithm>

#include "trackingobject.h"
#include "config.h"

// utility function to find median of a vector
int get_median(std::vector<int> &vec)
{
	std::sort(vec.begin(), vec.end());
	int median_value = 0;
	int median_count = 0;
	int this_count = 1;
	int last_val = INT_MAX;
	for(unsigned int i=0; i < vec.size(); i++)
	{
		// printf("val: %d, this_count: %d\n", vec.at(i), this_count);
		if (vec.at(i) == last_val)
			this_count++;
		else
			this_count = 1;

		if (this_count > median_count)
		{
			median_value = vec.at(i);
			median_count = this_count;
		}
		last_val = vec.at(i);
	}
	return median_value;
}

// constructor
//
// input image should be gray
// the trackingobject takes ownership over the image and will free() it when not needed anymore
// make sure you don't delete/change the *segments vector in other classes/functions, until finalize() was called
TrackingObject::TrackingObject(IplImage *image, CvPoint2D32f* points, char* status, int num)
{
	finalized = false;
	gray_image = image;
	active_segments = new std::set<Component*>;
	num_features = 0; // this variable is later set by addFeaturePoints

	center = cvPoint2D32f(0.0, 0.0);
	direction = cvPoint2D32f(0.0, 0.0);

	if (points && status && num)
		addFeaturePoints(points, status, num);
	else
		memset(this->status, FLOW_STATUS_LOST, MAX_FEATURE_POINTS);
}

// destructor
TrackingObject::~TrackingObject()
{
	//printf("Releasing tracking object\n");
	cvReleaseImage(&gray_image);
	delete active_segments; //FIXME
}

// copy constructor
TrackingObject::TrackingObject(const TrackingObject &pts)
{
	printf("cp constructor of TO called: %d\n", pts.num_features);
	gray_image = cvCloneImage(pts.gray_image);
	memcpy(features, pts.features, sizeof(CvPoint2D32f) * pts.num_features);
	memcpy(status, pts.status, sizeof(char) * pts.num_features);
	num_features = pts.num_features;

	center = pts.center;
	direction = pts.direction;
	finalized = true;
}

// checks that one segment only gets added once
void TrackingObject::addSegment(Component* comp)
{
	if( finalized )
	{
		fprintf(stderr, "Warning: Object finalized. Not adding segments\n");
		return;
	}

	if(!comp)
	{
		fprintf(stderr, "Warning: Not adding NULL component in addSegment()\n");
		return;
	}
	
	active_segments->insert(comp);
}

CvRect TrackingObject::getBoundingBox()
{
	if(finalized)
		return bbox;

	bbox = cvRect(-1, -1, 0, 0);
	if(active_segments->empty())
		return bbox;

	int minx,miny,maxx,maxy;

	std::set<Component*>::const_iterator it = active_segments->begin();
	Component *comp = *it;
	minx = comp->min_x;
	miny = comp->min_y;
	maxx = comp->max_x;
	maxy = comp->max_y;
	
	for(++it ; it != active_segments->end(); it++)
	{
		comp = *it;
		if (comp->min_x < minx) minx = comp->min_x;
		if (comp->min_y < miny) miny = comp->min_y;
		if (comp->max_x > maxx) maxx = comp->max_x;
		if (comp->max_y > maxy) maxy = comp->max_y;
	}
	bbox = cvRect(minx,miny, maxx-minx, maxy-miny);
	return bbox;
}

// clears current feature points, and finds new one for current segments
void TrackingObject::addFeaturePoints()
{
	if( finalized )
	{
		fprintf(stderr, "Warning: Object finalized. Not adding features\n");
		return;
	}

	// if we don't have any feature points yet, add MAX_FEATURE_POINTS
	if(num_features == 0)
		num_features = MAX_FEATURE_POINTS;

	// 1. find free "slots" in our feature point array which have bad status
	int free_fp = 0;
	for (int i=0; i<num_features; i++)
	{
		if(status[i] != FLOW_STATUS_OK)
			free_fp++;
	}

	// 2. calculate total pixels of all segments
	int total_pixels = 0;
	Component *comp = NULL;
	std::set<Component*>::const_iterator it;
	for(it = active_segments->begin(); it != active_segments->end(); it++)
	{
		comp = *it;
		total_pixels += comp->size;
	}
	printf("total: %d\n", total_pixels);

	// 3. distribute feature points accordingly
	for(it = active_segments->begin(); it != active_segments->end(); it++)
	{
		comp = *it;
		int fp = (int)(float(comp->size)/total_pixels * free_fp);
		//printf("fp: %d, free_fp: %d\n", fp, free_fp);
		comp->findGoodFeaturePoints(gray_image, features, status, fp);
	}

}

// this function is also responsible to calculate active segments
void TrackingObject::addFeaturePoints(CvPoint2D32f* points, char* status, int num)
{
	if( finalized )
	{
		fprintf(stderr, "Warning: Object finalized. Not adding features\n");
		return;
	}

	if (num + num_features > MAX_FEATURE_POINTS)
		num = MAX_FEATURE_POINTS - num_features;

	memcpy(features + num_features, points, sizeof(CvPoint2D32f) * num);
	memcpy(this->status + num_features, status, sizeof(char) * num);
	num_features += num;
}

void TrackingObject::finalize(TrackingObject* prev)
{
	getCenter();
	getDirection(prev);
	getBoundingBox();

	finalized = true;

	printf("Object finalized with %d active segment(s):\n"
		   "  center:    %4.2f ,%4.2f\n"
		   "  direction: %4.2f, %4.2f\n"
		   "  bounding box: %d,%d,%d,%d\n"
		   "  %d features total, %d found, %d predicted, %d lost\n",
		   active_segments->size(), center.x, center.y,
		   direction.x, direction.y, bbox.x, bbox.y, bbox.x+bbox.width, bbox.y+bbox.height,
		   num_features, num_found_features, num_predicted_features, num_lost_features);
}

// also stores number of found/lost/etc. features
CvPoint2D32f TrackingObject::getCenter()
{
	if(finalized)
		return center;

	center = cvPoint2D32f(0.0, 0.0);

	num_found_features = num_predicted_features = num_lost_features = 0;
	for(int i=0; i < num_features; i++) 
	{
		if (status[i] == FLOW_STATUS_OK)
			num_found_features++;

		// calculate center points from either found or estimated points
		if (status[i] >= FLOW_STATUS_OK && status[i] < FLOW_STATUS_LOST)
		{
			if(status[i] != FLOW_STATUS_OK)
				num_predicted_features++;

			center.x += features[i].x;
			center.y += features[i].y;
		}
		else
			num_lost_features++;
	}

	int num = num_found_features + num_predicted_features;
	if (!num)
		return center;

	center.x /= num;
	center.y /= num;
	return center;
}

// call with NULL to get last finalized value ???
CvPoint2D32f TrackingObject::getDirection(TrackingObject* prev)
{
	// printf("getDIR: %p: %d - %f,%f\n", prev, finalized, direction.x, direction.y);
	if (finalized)
		return direction;

	direction = cvPoint2D32f(0.0, 0.0);
	if( !prev )
		return direction;

	if (prev->num_features != num_features)
	{
		fprintf(stderr, "Error: Number of features in direction() unequal - %d, %d. Aborting...\n", prev->num_features, num_features);
		exit(EXIT_FAILURE);
	}

	int num = 0;
	for(int i=0; i < num_features; i++) 
	{
		if (status[i] == FLOW_STATUS_OK)
		{
			float x = features[i].x - prev->features[i].x;
			float y = features[i].y - prev->features[i].y;

			direction.x += x;
			direction.y += y;
			num++;
		}
	}

	if (!num)
	{
		direction = prev->getDirection();
		return direction;
	}

	direction.x /= num;
	direction.y /= num;

	return direction;
}
